/*
** path.c
**
** Copyright (c) 1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>

#include "pftpd.h"


char *user_ftp_dir = NULL;
char *server_ftp_dir = NULL;

/*
** Validate, merge CWD+ARG and trim virtual path
*/
char *
path_mk(FTPCLIENT *fp,
	const char *arg,
	char *buf,
	size_t bufsize)
{
    char *path, *cp;
    int alen, clen;


    if (arg == NULL)
	return NULL;

    if (debug > 2)
	fprintf(stderr, "path_mk, at start: cwd=%s, arg=%s\n",
		fp->cwd, arg);

    alen = strlen(arg);
    
    if (arg[0] == '/')
    {
	if (buf)
	{
	    if (alen+1 > bufsize)
		return NULL;

	    path = buf;
	}
	else
	    path = s_malloc(alen+1);
	
	strcpy(path, arg);
    }
    else
    {
	clen = strlen(fp->cwd);

	if (buf)
	{
	    if (clen+alen+1 > bufsize)
		return NULL;
	    path = buf;
	}
	else
	    path = s_malloc(clen+alen+1);
	
	strcpy(path, fp->cwd);
	if (clen > 0 && path[clen-1] != '/')
	    path[clen++] = '/';
	
	strcpy(path+clen, arg);
    }

    if (debug > 2)
	fprintf(stderr, "path_mk, after merge: path=%s\n", path);
    
    /* Trim any '..' components from the path */
    cp = path;
    while (*cp)
    {
	if (cp[0] == '/' &&
	    ((cp > path && cp[1] == '\0') || cp[1] == '/'))
	{
	    strcpy(cp, cp+1);
	}
	else if (cp[0] == '/' &&
		 cp[1] == '.' &&
		 (cp[2] == '\0' || cp[2] == '/'))
	{
	    strcpy(cp, cp+2);
	}
	else if (cp[0] == '/' &&
		 cp[1] == '.' &&
		 cp[2] == '.' &&
		 (cp[3] == '\0' || cp[3] == '/'))
	{
	    char *rest = cp+3;
	    while (cp > path && *--cp != '/')
		;
	    strcpy(cp, rest);
	}
	else
	    ++cp;
    }

    
    if (*path == '\0')
	strcpy(path, "/");
    
    if (debug > 2)
	fprintf(stderr, "path_mk, after trim: path=%s\n", path);
    
    return path;
}


char *
path_v2r(const char *path,
	 char *buf,
	 size_t bufsize)
{
    int blen, plen, dlen;


    if (path == NULL)
	return NULL;

    if (user_ftp_dir &&
	path[0] == '/' &&
	path[1] == '~')
    {
	/* User's public FTP dir */
	struct passwd pwb, *pwp = NULL;
	char user[128];
	char tmpbuf[2048], *cp;
	int err, len;
	

	cp = strchr(path+2, '/');
	if (cp)
	{
	    len = cp - (path+2);
	    memcpy(user, path+2, len);
	    user[len] = '\0';
	}
	else
	{
	    strcpy(user, path+2);
	    cp = "";
	}

	if (debug)
	    fprintf(stderr, "[%s]\n", user);
	
	err = s_getpwnam_r(user, &pwb, tmpbuf, sizeof(tmpbuf), &pwp);
	if (err || pwp == NULL)
	    return NULL;

	blen = strlen(pwp->pw_dir);
	dlen = strlen(user_ftp_dir);
	plen = strlen(path);
	
	if (buf)
	{
	    if (blen+dlen+plen+2 > bufsize)
		return NULL;
	}
	else
	    buf = s_malloc(blen+dlen+plen+2);
	
	strcpy(buf, pwp->pw_dir);
	buf[blen] = '/';
	strcpy(buf+blen+1, user_ftp_dir);
	strcpy(buf+dlen+blen+1, cp);

	if (debug)
	    fprintf(stderr, "path_v2r: %s -> %s\n", path, buf);
	
	return buf;
    }

    blen = strlen(server_ftp_dir);
    plen = strlen(path);

    if (buf)
    {
	if (blen+plen+1 > bufsize)
	    return NULL;
    }
    else
	buf = s_malloc(blen+plen+1);
    
    strcpy(buf, server_ftp_dir);
    strcpy(buf+blen, path);

    if (debug)
	fprintf(stderr, "path_v2r: %s -> %s\n", path, buf);
	
    return buf;
}


